

"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class FillNaTransformer_median(BaseEstimator, TransformerMixin):

    """This transformer handles missing values.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_median/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))

        self.train_median = X[self.columns].median()

        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "train_median"):
            raise AttributeError("FillNaTransformer_median has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X[self.columns] = X[self.columns].fillna(self.train_median)

        return X


class FillNaTransformer_mean(BaseEstimator, TransformerMixin):

    """This transformer handles missing values.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_mean/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))

        self.train_mean = X[self.columns].mean()

        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "train_mean"):
            raise AttributeError("FillNaTransformer_median has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X[self.columns] = X[self.columns].fillna(self.train_mean)

        return X


class FillNaTransformer_idmax(BaseEstimator, TransformerMixin):

    """This transformer handles missing values for idmax.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_idmax/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            # self.columns = X.select_dtypes(exclude=["object"]).columnsX.select_dtypes(exclude=["object"]).columns
            self.columns = X.columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))

        self._fitted = True

        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "_fitted"):
            raise AttributeError("FillNaTransformer_idmax has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        for col in self.columns:
            X[col] = X[col].fillna(X[col].value_counts().idxmax())

        return X


class FillNaTransformer_any(BaseEstimator, TransformerMixin):

    """This transformer delete row that there is some NaN.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_any/

    """
    def __init__(self):
        """Init replace missing values."""

    def fit(self, X, y=None, **fit_params):
        """Not implemented.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        return self

    def transform(self, X):
        """This transformer delete row that there is some NaN

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        self.columns = X.columns
        X.dropna(axis=0, how="any", inplace=True)
        # X.reset_index()
        # X.drop(["index"], axis=1)

        return X


class FillNaTransformer_all(BaseEstimator, TransformerMixin):

    """This transformer delete row that there is all NaN.

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_all/

    """
    def __init__(self):
        """Init replace missing values."""

    def fit(self, X, y=None, **fit_params):
        """Not implemented.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        return self

    def transform(self, X):
        """This transformer delete row that there is some NaN

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        self.columns = X.columns
        X.dropna(axis=0, how="all", inplace=True)
        # X.reset_index()
        # X.drop(["index"], axis=1, inplace=True)

        return X


class FillNaTransformer_value(BaseEstimator, TransformerMixin):

    """This transformer handles missing values.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_value/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, value=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if value is not None:
            if (not isinstance(value, bool) and not isinstance(value, int)
                    and not isinstance(value, float) and not isinstance(value, str)):
                raise NameError("Type value not permited: {}".format(type(value)))
            else:
                self.value = value
        else:
            raise NameError("Not alowed value=None")

        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))

        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "value"):
            raise AttributeError("FillNaTransformer_value has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X[self.columns] = X[self.columns].fillna(self.value)
        return X


class FillNaTransformer_backward(BaseEstimator, TransformerMixin):

    """This transformer handles missing values closer backward.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_backward/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))
        self._fitted = True
        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "_fitted"):
            raise AttributeError("FillNaTransformer_backward has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))
        X_transform = X.copy()
        X_transform[self.columns] = X[self.columns].fillna(method="bfill")
        return X_transform


class FillNaTransformer_forward(BaseEstimator, TransformerMixin):

    """This transformer handles missing values closer forward.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FillNaTransformer_forward/

    """
    def __init__(self, columns=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))
        self._fitted = True
        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "_fitted"):
            raise AttributeError("FillNaTransformer_backward has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        X_transform[self.columns] = X[self.columns].fillna(method="ffill")
        return X_transform



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import os
import matplotlib.pyplot as plt

import tensorflow as tf


def load_image(filename):

    return tf.cast(tf.image.decode_png(
                        tf.io.read_file(filename), channels=1), tf.float32)


def plot_image(img, name="Image", save=False, logdir_report="/images"):
    """Grafico Imagen."""
    plt.imshow(img[:, :, 0], cmap='gray', interpolation='none')
    plt.title(name)
    plt.axis('off')
    plt.show()
    if save:
        name = name + ".png"
        filename = os.path.join(logdir_report, name)
        plt.savefig(filename)
        plt.close()
    plt.show()


def plot_image2(img1, img2, title="Images", save=False, logdir_report="/images"):
    """Grafico Imagen."""
    fig, axs = plt.subplots(1, 2)
    ax = axs.flatten()
    fig.suptitle(title, fontsize=20)

    ax[0].axis('off')
    ax[0].imshow(img1[:, :, 0], cmap='gray', interpolation='none')

    ax[1].axis('off')
    ax[1].imshow(img2[:, :, 0], cmap='gray', interpolation='none')

    plt.show()
    if save:
        name = title + ".png"
        filename = os.path.join(logdir_report, name)
        fig.savefig(filename)
        fig.close()
    plt.show()




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT


Machine Learning for Data engineer
==================================

MLearner is Python module based on scikit-learn.
It is based on high-level libraries for data engineers.


See https://jaisenbe58r.github.io/MLearner/ for complete documentation.

"""

import logging
import os

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
#   X.Y
#   X.Y.Z   # For bugfix releases
#
# Admissible pre-release markers:
#   X.YaN   # Alpha release
#   X.YbN   # Beta release
#   X.YrcN  # Release Candidate
#   X.Y     # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#
__version__ = '0.2.2'


# On OSX, we can get a runtime error due to multiple OpenMP libraries loaded
# simultaneously. This can happen for instance when calling BLAS inside a
# prange. Setting the following environment variable allows multiple OpenMP
# libraries to be loaded. It should not degrade performances since we manually
# take care of potential over-subcription performance issues, in sections of
# the code where nested OpenMP loops can happen, by dynamically reconfiguring
# the inner OpenMP runtime to temporarily disable it while under the scope of
# the outer OpenMP parallel section.
os.environ.setdefault("KMP_DUPLICATE_LIB_OK", "True")

# Workaround issue discovered in intel-openmp 2019.5:
# https://github.com/ContinuumIO/anaconda-issues/issues/11294
os.environ.setdefault("KMP_INIT_AT_FORK", "FALSE")

__all__ = ["classifier", "data", "evaluation", "externals",
            "feature_selection", "load", "images", "models", "neural", "preprocessing", "plotly",
            "training", "utils", "stacking"]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import pickle

from sklearn.metrics import roc_curve, auc, classification_report, confusion_matrix


class EvaluationModels(object):
    """
    Evaluacion de modelosself.

    Parameters
    ----------
    model : Modelo a evaluearself.
    random_state : Asignacion de la semilla de aleatoriedad.

    Attributes
    ----------


    See Also
    --------
    mlearner.load.DataLoad : Loading a dataset from a file or dataframe.

    """

    def __init__(self, model, random_state=99):
        """
        Inicializacion e la clase de Preprocesado de un dataframe
        """
        self.model = model
        self.random_state = random_state

    def add_model(self, filename):
        """
        Load the model from disk
        """
        self.model = pickle.load(open(filename, 'rb'))

    def restore_model(self, filename):
        """
        Load the model from disk
        """
        self.model = pickle.load(open(filename, 'rb'))

    def save_model(self, filename):
        """
        Save the model to disk
        """
        pickle.dump(self.model, open(filename, 'wb'))

    def confusion_matrix(self, y_true, y_pred):
        from sklearn.metrics import confusion_matrix
        cm = confusion_matrix(y_true, y_pred)
        return cm

    def validacion_cruzada(self, X, Y, n_splits, shuffle=True, scoring="accuracy"):
        """
        Validacion cruzada del dataset introducido como input.

        Inputs:
            - cv = Numero de iteraciones.

        Outputs:
            - score: Media de los acc de todas las iteraciones.
        """
        from sklearn.model_selection import cross_val_score
        from sklearn.model_selection import KFold

        cv = KFold(n_splits=n_splits, shuffle=True, random_state=self.random_state)
        scores = cross_val_score(self.model, X, Y, scoring=scoring, cv=cv)

        return scores.mean()

    def evaluacion_rf_2features(self, clf, data_eval, data_eval_target, targets=[0, 1], save=False, logdir_report="", display=True):
        """
        Funcion que nos selecciona el thresholder mas optimo:

        Inputs:

            - data_eval: Conjunto de evaluacion.
            - probs: Matriz de probabilidades que viene dado por el "lm.predict_proba" del podelo de RL.
            - thresholders: Intervalo de corte para determinar la categoria. viene dado por el "metrics.roc_curve" de sckitlearn.
            - display: si queremos mostrar los datos graficados por pantalla.

        Outputs:

            - th: Thresholder seleccionado.
            - res: El accurancy con ese thresholder.
            - df: Dataframe con el resultado del test.
        """
        results = []

        probs = clf.predict_proba(data_eval)
        prob = probs[:, 1]
        df = pd.DataFrame(prob)
        df = df.set_index(np.array(list(data_eval.index)))

        data_remp = data_eval_target.copy()
        data_remp = data_remp.replace({targets[1]: 1, targets[0]: 0})

        _, _, thresholders = roc_curve(data_remp, prob)

        for i in range(len(thresholders)):

            df["prediction"] = np.where(df[0] >= thresholders[i], targets[1], targets[0])
            df["actual"] = list(data_eval_target)
            df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

            resultado = df["result"].sum()/df["result"].count()
            results.append(resultado)

        index = results.index(max(results))
        th = thresholders[index]
        res = max(results)

        if display:

            plt.plot(thresholders, results, marker="o", linestyle="--", color="r")
            x = [th, th]
            y = [0, 1]
            plt.plot(x, y)
            plt.xlabel("thresholders")
            plt.ylabel("Result")
            plt.title("Resultado segun thresholders")
            plt.xlim(-0.1, 1.1)
            plt.show()
            if save:
                plt.savefig(logdir_report + "_eval.png")

        df["prediction"] = np.where(df[0] >= thresholders[index], targets[1], targets[0])
        df["actual"] = list(data_eval_target)
        df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

        return th, res, df

    def create_ROC(self, lm, X_test, Y_test, targets=[0, 1], logdir_report="", display=True, save_image=True):
        """
        Se crea la curva ROC de las predicciones de conjunto de test.

        Outputs:
            - df: Dataframe de los datos de metricas ROC.
            - auc: Area por debajo de la curfa ROC (efectividad de las predicciones).
        """
        probs = lm.predict_proba(X_test)
        prob = probs[:, 1]

        data_remp = Y_test.copy()
        data_remp = data_remp.replace({targets[1]: 1, targets[0]: 0})

        espc_1, sensit, thresholds = roc_curve(data_remp, prob)

        df = pd.DataFrame({
            "esp": espc_1,
            "sens": sensit,
            "TH": thresholds
        })

        _auc = auc(espc_1, sensit)

        if display:

            plt.plot(df.esp, df.sens, marker="o", linestyle="--", color="r")
            x = [i*0.01 for i in range(100)]
            y = [i*0.01 for i in range(100)]
            plt.plot(x, y)
            plt.xlabel("1-Especifidad")
            plt.ylabel("Sensibilidad")
            plt.title("Curva ROC")

            if save_image:
                plt.savefig(logdir_report + "_ROC.png")

            # plt.close()

        return df, _auc

    def create_ROC_pro(self, fprs, tprs, X, y, targets=[0, 1], logdir_report="", display=True, save_image=True):
        """
        Plot the Receiver Operating Characteristic from a list
        of true positive rates and false positive rates.
        """
        # Initialize useful lists + the plot axes.
        tprs_interp = []
        aucs = []
        mean_fpr = np.linspace(0, 1, 100)
        f, ax = plt.subplots(figsize=(14, 10))

        # remplazo de datos a [0, 1]
        data_remp = y.copy()
        y = data_remp.replace({targets[1]: 1, targets[0]: 0})

        # Plot ROC for each K-Fold + compute AUC scores.
        for i, (fpr, tpr) in enumerate(zip(fprs, tprs)):
            tprs_interp.append(np.interp(mean_fpr, fpr, tpr))
            tprs_interp[-1][0] = 0.0
            roc_auc = auc(fpr, tpr)
            aucs.append(roc_auc)
            ax.plot(fpr, tpr, lw=1, alpha=0.3,
                    label='ROC fold %d (AUC = %0.2f)' % (i+1, roc_auc))

        # Plot the luck line.
        plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r', label='Luck', alpha=.8)

        # Plot the mean ROC.
        mean_tpr = np.mean(tprs_interp, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        ax.plot(mean_fpr, mean_tpr, color='b',
                label=r'Mean ROC (AUC = %0.2f $\pm$ %0.2f)' % (mean_auc, std_auc),
                lw=2, alpha=.8)

        # Plot the standard deviation around the mean ROC.
        std_tpr = np.std(tprs_interp, axis=0)
        tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
        tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax.fill_between(mean_fpr, tprs_lower, tprs_upper, color='grey', alpha=.2,
                        label=r'$\pm$ 1 std. dev.')

        # Fine tune and show the plot.
        ax.set_xlim([-0.05, 1.05])
        ax.set_ylim([-0.05, 1.05])
        ax.set_xlabel('False Positive Rate')
        ax.set_ylabel('True Positive Rate')
        ax.set_title('Receiver operating characteristic')
        ax.legend(loc="lower right")

        if display:
            plt.show()
        if save_image:
            plt.savefig(logdir_report + "_ROC_pro.png")

            plt.close()

        return (f, ax)

    def compute_roc_auc(self, model, index, X, y):
        """
        Computo para todas las pruebas de KFold
        """
        y_predict = model.predict_proba(X.iloc[index])[:, 1]
        fpr, tpr, _ = roc_curve(y.iloc[index], y_predict)
        auc_score = auc(fpr, tpr)

        return fpr, tpr, auc_score

    def class_report(self, y_true, predictions, clases, save=False, logdir_report=""):
        """
        Un informe de clasificacion se utiliza para medir la calidad de las predicciones de un
        algoritmo de clasificacion. Cuantas predicciones son verdaderas y cuantas son falsas.
        Mas especificamente, los Positivos verdaderos, los Positivos falsos, los negativos
        verdaderos y los negativos falsos se utilizan para predecir las metricas de un informe
        de clasificacion.

        El informe muestra las principales metricas de clasificacion de precision, recuperacion
        y puntaje f1 por clase. Las metricas se calculan utilizando verdaderos y falsos positivos,
        verdaderos y falsos negativos. Positivo y negativo en este caso son nombres genericos
        para las clases predichas. Hay cuatro formas de verificar si las predicciones son
        correctas o incorrectas:

            TN / Verdadero negativo: cuando un caso fue negativo y se pronostico negativo
            TP / Verdadero Positivo: cuando un caso fue positivo y predicho positivo
            FN / Falso negativo: cuando un caso fue positivo pero predicho negativo
            FP / Falso Positivo: cuando un caso fue negativo pero predicho positivo

        La precision es la capacidad de un clasificador de no etiquetar una instancia positiva
        que en realidad es negativa. Para cada clase se define como la relacion de positivos
        verdaderos a la suma de positivos verdaderos y falsos.

            TP - Positivos verdaderos
            FP - Positivos falsos

            Precision: precision de las predicciones positivas.
            Precision = TP / (TP + FP)

        Recordar es la capacidad de un clasificador para encontrar todas las instancias positivas.
        Para cada clase se define como la relacion entre los verdaderos positivos y la suma de
        los verdaderos positivos y los falsos negativos.

            FN - Falsos negativos

            Recordar: fraccion de positivos identificados correctamente.
            Recuperacion = TP / (TP + FN)

        El  puntaje F 1 es una media armonica ponderada de precision y recuperacion de modo que
        el mejor puntaje es 1.0 y el peor es 0.0. En terminos generales, los  puntajes de F 1 son
        mas bajos que las medidas de precision, ya que incorporan precision y recuerdo en su
        calculo. Como regla general, el promedio ponderado de F 1  debe usarse para comparar
        modelos clasificadores, no la precision global.

            Puntuacion F1 = 2 * (recuperacion * precision) / (recuperacion + precision)
        """
        target_names = ["{}".format(clases[i]) for i in range(len(clases))]
        cr = classification_report(y_true, predictions, target_names=target_names)
        print("\nInforme de clasificacion:\n")
        print(cr)
        if save:
            print(cr, file=open(logdir_report + "_Informe_Clasificacion" + ".txt", 'w'))

        return cr

    def plot_Histograma(self, predict, correct, incorrect, logdir_report, categorias=[0, 1],
                        display=True, save_image=True):
        import tensorflow as tf
        i_pred_correct = tf.argmax(predict[correct], axis=1, output_type=tf.int32)
        i_pred_incorrect = tf.argmax(predict[incorrect], axis=1, output_type=tf.int32)

        plt.figure()
        _, ax = plt.subplots(2, 1)

        ax[0].set_ylabel('frequencia')
        ax[0].set_xlabel('valores')
        ax[0].legend(['Seeds', 'Craters'])
        ax[0].set_title('Histograma - Predicciones correctas')

        ax[1].set_ylabel('frequencia')
        ax[1].set_xlabel('valores')
        ax[1].legend(['Seeds', 'Craters'])
        ax[1].set_title('Histograma - Predicciones incorrectas')

        for i in categorias:
            i_pred_correct = np.nonzero(i_pred_correct == i)[0]
            ax[0].hist(predict[i_pred_correct][i_pred_correct, 0], alpha=0.8)
            i_pred_incorrect = np.nonzero(i_pred_incorrect == i)[0]
            ax[1].hist(predict[incorrect][i_pred_incorrect, 0], alpha=0.8)

        plt.tight_layout()

        if display:
            plt.show()

        if save_image:
            plt.savefig(logdir_report + "_histogram.png")

    def plot_confusion_matrix(self, y_true, y_pred, classes, num_clases, logdir_report,
                                normalize=False, title=None, cmap=cm.Blues, name="cm_normalizada"):
        """
        Una matriz de confusion es un resumen de los resultados de prediccion sobre un problema
        de clasificacion.

        El numero de predicciones correctas e incorrectas se resume con valores de conteo y se
        desglosa por clase. Esta es la clave de la matriz de confusion.

        La matriz de confusion muestra las formas en que su modelo de clasificacion
        se confunde cuando hace predicciones.

        Le da una idea no solo de los errores que esta cometiendo su clasificador, sino mas
        importante aun, de los tipos de errores que se estan cometiendo.

        Es este desglose el que supera la limitacion del uso de la precision de clasificacion solo.
        """

        """
        This function prints and plots the confusion matrix.
        Normalization can be applied by setting `normalize=True`.
        """
        if not title:
            if normalize:
                title = "Matriz de confusion normalizada"
            else:
                title = "Matriz de confusion sin normalizar"
        # Compute confusion matrix
        cm = confusion_matrix(y_true, y_pred)
        # Only use the labels that appear in the data
        classes = classes[0:num_clases]
        if normalize:
            cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]
            print("\nMatriz de confusion normalizada:\n")
        else:
            print("\nMatriz de confusion sin normalizar:\n")

        print(cm)

        fig, ax = plt.subplots()

        im = ax.imshow(cm, interpolation="nearest", cmap=cmap)
        ax.figure.colorbar(im, ax=ax)
        # We want to show all ticks...
        ax.set(
            xticks=np.arange(cm.shape[1]),
            yticks=np.arange(cm.shape[0]),
            # ... and label them with the respective list entries
            xticklabels=classes,
            yticklabels=classes,
            title=title,
            ylabel="Clase real",
            xlabel="Clase predecida",
        )

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), ha="right", rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        fmt = ".2f" if normalize else "d"
        thresh = cm.max() / 2.0
        for i in range(cm.shape[0]):
            for j in range(cm.shape[1]):
                ax.text(
                    j,
                    i,
                    format(cm[i, j], fmt),
                    horizontalalignment="center",
                    verticalalignment="center",
                    color="white" if cm[i, j] > thresh else "black",
                )
        fig.tight_layout()
        plt.savefig(logdir_report + "_" + name + "_cm.png")
        # plt.show()
        plt.close()
        return ax




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer <www.linkedin.com/in/jaisenbe>
License: MIT
"""

from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.externals import six
from sklearn.base import clone
from sklearn.pipeline import _name_estimators
import numpy as np
import operator


class MajorityVoteClassifier(BaseEstimator, 
                             ClassifierMixin):
    """ A majority vote ensemble classifier

    Parameters
    ----------
    classifiers : array-like, shape = [n_classifiers]
      Different classifiers for the ensemble

    vote : str, {'classlabel', 'probability'} (default='label')
      If 'classlabel' the prediction is based on the argmax of
        class labels. Else if 'probability', the argmax of
        the sum of probabilities is used to predict the class label
        (recommended for calibrated classifiers).

    weights : array-like, shape = [n_classifiers], optional (default=None)
      If a list of `int` or `float` values are provided, the classifiers
      are weighted by importance; Uses uniform weights if `weights=None`.

    """
    def __init__(self, classifiers, vote='classlabel', weights=None):

        self.classifiers = classifiers
        self.named_classifiers = {key: value for key, value
                                  in _name_estimators(classifiers)}
        self.vote = vote
        self.weights = weights

    def fit(self, X, y):
        """ Fit classifiers.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Matrix of training samples.

        y : array-like, shape = [n_samples]
            Vector of target class labels.

        Returns
        -------
        self : object

        """
        if self.vote not in ('probability', 'classlabel'):
            raise ValueError("vote must be 'probability' or 'classlabel'"
                             "; got (vote=%r)"
                             % self.vote)

        if self.weights and len(self.weights) != len(self.classifiers):
            raise ValueError('Number of classifiers and weights must be equal'
                             '; got %d weights, %d classifiers'
                             % (len(self.weights), len(self.classifiers)))

        # Use LabelEncoder to ensure class labels start with 0, which
        # is important for np.argmax call in self.predict
        self.lablenc_ = LabelEncoder()
        self.lablenc_.fit(y)
        self.classes_ = self.lablenc_.classes_
        self.classifiers_ = []
        for clf in self.classifiers:
            fitted_clf = clone(clf).fit(X, self.lablenc_.transform(y))
            self.classifiers_.append(fitted_clf)
        return self

    def predict(self, X):
        """ Predict class labels for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Matrix of training samples.

        Returns
        ----------
        maj_vote : array-like, shape = [n_samples]
            Predicted class labels.
            
        """
        if self.vote == 'probability':
            maj_vote = np.argmax(self.predict_proba(X), axis=1)
        else:  # 'classlabel' vote

            #  Collect results from clf.predict calls
            predictions = np.asarray([clf.predict(X)
                                      for clf in self.classifiers_]).T

            maj_vote = np.apply_along_axis(
                                      lambda x:
                                      np.argmax(np.bincount(x,
                                                weights=self.weights)),
                                      axis=1,
                                      arr=predictions)
        maj_vote = self.lablenc_.inverse_transform(maj_vote)
        return maj_vote

    def predict_proba(self, X):
        """ Predict class probabilities for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            Training vectors, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        ----------
        avg_proba : array-like, shape = [n_samples, n_classes]
            Weighted average probability for each class per sample.

        """
        probas = np.asarray([clf.predict_proba(X)
                             for clf in self.classifiers_])
        avg_proba = np.average(probas, axis=0, weights=self.weights)
        return avg_proba

    def get_params(self, deep=True):
        """ Get classifier parameter names for GridSearch"""
        if not deep:
            return super(MajorityVoteClassifier, self).get_params(deep=False)
        else:
            out = self.named_classifiers.copy()
            for name, step in six.iteritems(self.named_classifiers):
                for key, value in six.iteritems(step.get_params(deep=True)):
                    out['%s__%s' % (name, key)] = value
            return out


"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

class OrientationClassTransformer(BaseEstimator, TransformerMixin):
    """Transformer Orientation."""
    def __init__(self, columns, name="OCT", a=135, b=45):
        self.columns = columns
        self.name = name
        self.a, self.b = a, b

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        func = lambda x: 0 if (x <= self.a and x >= self.b) else 1
        d = map(func, X[self.columns].values)

        return np.fromiter(d, dtype=np.int32).reshape(-1, 1)


class MFD_OrientationClassTransformer(BaseEstimator, TransformerMixin):
    """Transformer MFD Orientation."""
    def __init__(self, columns, name="MFDOCT", a=120, b=60, c=30, d=150):
        self.columns = columns
        self.name = name
        self.a, self.b = a, b
        self.c, self.d = c, d

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        func = lambda x: 0 if (x <= self.a and x >= self.b) else ( 1 if (x < self.c or x > self.d) else 2)
        d = map(func, X[self.columns].values)

        return np.fromiter(d, dtype=np.int32).reshape(-1, 1)


class ClassTransformer_value(BaseEstimator, TransformerMixin):
    def __init__(self, columns, name="A/AH_cat", value=100):
        self.columns = columns
        self.name = name
        self.value = value

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        func = lambda x: 0 if x == self.value else 1
        d = map(func, X[self.columns].values)

        return np.fromiter(d, dtype=np.int32).reshape(-1, 1)


class Keep(BaseEstimator, TransformerMixin):
    """Mantener columnas."""
    def __init__(self):
        pass
    def fit(self, X, y=None):
        return self
    def transform(self, X):
        return X


"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .ml_clasifier import TrainingUtilities, PipelineClasificators, wrapper_model, wrapper_pipeline


__all__ = ["TrainingUtilities", "PipelineClasificators", "wrapper_model", "wrapper_pipeline"]
